(*********************************************************************

  HOC02BaseLemmas

**********************************************************************

  Copyright (C) 2014
  Simon Boulier, Martin Escarra, Petar Maksimovic, Alan Schmitt

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

**********************************************************************)

Require Import HOC01Defs.

(**********************************
  PREPARING THE CANONICAL
 **********************************)

(* The LVariables which have in their scope the GVariable X *)
Fixpoint E (X:var) (p:process) :=
  match p with
    | Send _ p      => (E X p)
    | Receive _ x p => ifb (X#p)
                         then \{} 
                         else (\{x} \u (E X p))
    | Lvar _        => \{}
    | Gvar _        => \{}
    | Par p q       => (E X p) \u (E X q)
    | Nil           => \{}
  end.

(**********************************
  BASIC FRESHNESS
 **********************************)

(********** GVAR FRESHNESS **********)

(* If there are no GVariables in a process, any one is fresh *)
Lemma fresh_GV : forall p X, GV p = \{} -> X#p.
Proof. 
  unfolds fresh_gvar. introv H. rewrites~ H. 
Qed.

(* If a GVariable is fresh in a Send, it is also fresh in the sender *)
Lemma gfresh_send : forall a X p, X # (Send a p) = X # p.
Proof. 
  auto.
Qed.

(* GVariable freshness of reception implies GVariable freshness of recipient *)
Lemma gfresh_receive : forall a x X p, X # (Receive a x p) = X # p.
Proof. 
  auto.
Qed.

(* A GVariable is fresh in an LVariable *)
Lemma gfresh_lvar : forall x X, X # (Lvar x).
Proof.
  unfolds fresh_gvar. 
  intro. applys* in_empty_elim.
Qed.

(* Freshness and equality for GVariables *)
Lemma gfresh_gvar : forall X Y, X # (Gvar Y) = (X <> Y).
Proof.
  introv; unfold fresh_gvar; simpl.  
  unfold notin; rewrite in_singleton; tauto.
Qed.

(* GVariable freshness of parallelisation *)
Lemma gfresh_par_iff : forall X p1 p2,  
  X # (Par p1 p2) = (X#p1 /\ X#p2).
Proof.
  introv; unfolds fresh_gvar; simpl.
  unfold notin; rewrite in_union; rew_logic.
  tauto.
Qed.

(* A GVariable is fresh in Nil *)
Lemma gfresh_nil : forall X, X # Nil.
Proof.
  introv; unfolds fresh_gvar; simpls*.
Qed.

(********** LVAR FRESHNESS **********)

(* If there are no LVariables in a process, any one is fresh *)
Lemma fresh_LV : forall p x, LV p = \{} -> x ## p.
Proof. 
  unfolds fresh_lvar. introv H. rewrites~ H. 
Qed.

(* LVariable freshness in send *)
Lemma lfresh_send : forall a x p,
  x ## (Send a p) = x ## p.
Proof.
  auto.
Qed.

(* LVariable freshness in receive *)
Lemma lfresh_receive : forall a x y p,
  x ## (Receive a y p) = (x=y \/ x ## p).
Proof.
  introv; unfold fresh_lvar; simpl. 
  unfold notin; rewrite in_remove.
  unfold notin; rewrite in_singleton.
  rew_logic. tauto.
Qed.

(* Freshness and equality for LVariables *)
Lemma lfresh_lvar : forall x y,
  x ## (Lvar y) = (x <> y).
Proof.
  introv; unfolds fresh_lvar; simpl.
  unfold notin; rewrite~ in_singleton.
Qed.

(* An LVariable is fresh in an GVariable *)
Lemma lfresh_gvar : forall x X,
  x ## (Gvar X).
Proof.
  introv; unfold fresh_lvar; simpls~.
Qed.

(* GVariable freshness of parallelisation *)
Lemma lfresh_par_iff : forall x p1 p2,
  x ## (Par p1 p2) = (x##p1 /\ x##p2).
Proof.
  introv; unfold fresh_lvar; simpls~.
  unfold notin; rewrite in_union; rew_logic.
  tauto.
Qed.

(* An LVariable is fresh in Nil *)
Lemma lfresh_nil : forall x,
  x ## Nil.
Proof.
  unfolds fresh_lvar; introv; simpls~.
Qed.

(********** CHANNEL FRESHNESS **********)

(* Channel freshness in send *)
Lemma cfresh_send : forall a m p,
  m ### (Send a p) = (m ### p /\ a <> m).
Proof.
  introv; unfolds fresh_chan; simpls. 
  unfold notin; rewrite in_union; rewrite in_singleton. 
  rew_logic. intuition.
Qed.

(* Channel freshness in receive *)
Lemma cfresh_receive : forall a m y p,
  m ### (Receive a y p) = (m ### p /\ a <> m).
Proof.
  introv; unfolds fresh_chan; simpls*. 
  unfold notin; rewrite in_union; rewrite in_singleton. 
  rew_logic; intuition.
Qed.

(* A channel is fresh in an LVariable *)
Lemma cfresh_lvar : forall m y,
  m ### (Lvar y).
Proof.
  introv. unfolds. simpls~.
Qed.

(* A channel is fresh in a GVariable *)
Lemma cfresh_gvar :  forall m X,
  m ### (Gvar X).
Proof.
  introv. unfolds. simpls~.
Qed.

(* Channel freshness of parallelisation *)
Lemma cfresh_par_iff : forall m p1 p2,
  m ### (Par p1 p2) = (m###p1 /\ m###p2).
Proof.
  introv; unfold fresh_chan; simpls.
  unfold notin; rewrite in_union; rew_logic; tauto.
Qed.

(* A channel is fresh in Nil *)
Lemma cfresh_nil : forall m,
  m ### Nil.
Proof.
  introv. unfolds. simpls~.
Qed.

(* Channel freshness with channel transmitting messages *)
Lemma cfresh_send_nil: forall m n, m ### Send n Nil = (m <> n).
Proof.
  introv; unfold fresh_chan; simpls.
  rewrite union_empty_r; unfold notin; rewrite in_singleton; intuition.
Qed.

Ltac freshen_up :=
  try solve [rewrite gfresh_send in *; jauto |
             rewrite gfresh_receive in *; jauto |
             rewrite gfresh_gvar in *; jauto |
             rewrite gfresh_par_iff in *; try splits; jauto |
             rewrite lfresh_send in *; jauto |
             rewrite lfresh_receive in *; jauto |
             rewrite lfresh_lvar in *; jauto |
             rewrite lfresh_par_iff in *; try splits; jauto |
             rewrite cfresh_send in *; jauto |
             rewrite cfresh_receive in *; jauto |
             rewrite cfresh_par_iff in *; try splits; jauto |
             apply* gfresh_lvar | apply* gfresh_nil |
             apply* lfresh_gvar | apply* lfresh_nil |
             apply* cfresh_lvar | apply* cfresh_gvar | apply* cfresh_nil 
             ].

Ltac simpl_fresh :=
  try (rewrite gfresh_send in *);
  try (rewrite gfresh_receive in *);
  try (rewrite gfresh_gvar in *);
  try (rewrite gfresh_par_iff in *); 
  try (rewrite lfresh_send in *);
  try (rewrite lfresh_receive in *); 
  try (rewrite lfresh_lvar in *);
  try (rewrite lfresh_par_iff in *);
  try (rewrite cfresh_send in *);
  try (rewrite cfresh_receive in *); 
  try (rewrite cfresh_par_iff in *);
  try (rewrite cfresh_send_nil in *).

(**********************************
  BASIC SIZE LEMMAS
 **********************************)

(* Substitution with something of size one doesn't change the size *)
Lemma size_subst : forall X p q,
  (size q) = 1 -> (size ([q // X]p)) = (size p).
Proof.
  induction p; simpls*; cases_if*.
Qed.

Lemma size_lsubst : forall x p q,
  size q = 1 -> size p = size ([q /// x]p).
Proof.
  introv H.
  inductions p; simpls* ; cases_if* ; simpls*.
Qed.

(* Substitution with something of size one doesn't change the size *)
Lemma sizeP_subst : forall X p q,
  (sizeP q) = 1 -> (sizeP ([q // X]p)) = (sizeP p).
Proof.
  induction p; simpls*; cases_if*.
Qed. 

(**********************************
  BASIC SWAPPING LEMMAS
 **********************************)

(********** SWAPPING VARIABLES **********)

(* Swapping the left swappand  *)
Lemma swap_gvar_left : forall X X', {{X,X'}}X = X'.
Proof.
  introv. unfolds. cases_if*.
Qed.

(* Swapping the right swappand *)
Lemma swap_gvar_right : forall X X', {{X,X'}}X' = X.
Proof.
  introv. unfolds. cases_if*. cases_if*.
Qed. 

(* Swapping for itself in a GVariable *)
Lemma swap_gvar_equal : forall X Y,
  {{X,X}}Y = Y.
Proof.
  introv. unfolds ; cases_if* ; cases_if*.
Qed.

(* Swapping for itself in a process *)
Lemma swap_equal : forall X p, {{[X,X]}}p = p.
Proof.
  inductions p; simpls*; fequals. apply swap_gvar_equal.
Qed.

(* Swapping preserves equality of GVariables  *)
Lemma swap_gvar_equal_iff : forall X X' Y Y',
  Y = Y' <-> {{X,X'}}Y = {{X,X'}}Y'.
Proof.
  split. introv H. congruence.
  introv H. unfolds in H ; repeat cases_if*; repeat congruence.
Qed.

(* Swap is involutive on GVariables *)
Lemma swap_gvar_involutive : forall X X' Y,
  {{X,X'}}({{X,X'}}Y) = Y.
Proof.
  introv ; unfolds ; repeat cases_if*. congruence.
Qed.

(* Swap is involutive on processes *)
Lemma swap_involutive : forall X Y p,  {{[X,Y]}}({{[X,Y]}}p) = p.
Proof.
  inductions p; simpls*; fequals. apply swap_gvar_involutive.
Qed.


(* Swap is symmetric on processes *)
Lemma swap_sym : forall p X Y, {{[X,Y]}}p = {{[Y,X]}}p.
Proof.
  inductions p; introv; simpls; try rewrite~ IHp; auto.
  unfold swap_gvar; repeat cases_if*; substs~.
  rewrite IHp1; rewrite~ IHp2.
Qed.

(********** SWAPPING CHANNELS **********)

(* Swapping for itself in a name *)
Lemma swap_name_equal : forall m n,
  {{|m,m|}}n = n.
Proof.
  introv. unfolds ; cases_if* ; cases_if*.
Qed.

(* Swapping for itself in a process *)
Lemma swap_chan_equal : forall m p, |m,m|p = p.
Proof.
  inductions p; simpls*; fequals; apply swap_name_equal.
Qed.

(* Swap is symmetric on processes *)
Lemma swap_chan_sym : forall p m n, |m, n|p = |n, m|p.
Proof.
 inductions p; introv; simpls; try rewrite~ IHp; auto.
  unfold swap_name; repeat cases_if*; substs~.
  unfold swap_name; repeat cases_if*; substs~.
  rewrite IHp1; rewrite~ IHp2.
Qed.

(* Swapping channels is involutive on channels *)
Lemma swap_name_involutive : forall m n a,
  {{|m,n|}}({{|m,n|}}a) = a.
Proof.
  introv ; unfolds ; repeat cases_if*. congruence.
Qed.

(* Swapping channels is involutive on processes *)
Lemma swap_chan_involutive : forall m n p,
  |m,n|(|m,n|p) = p.
Proof.
  inductions p; simpls*; fequals; apply swap_name_involutive.
Qed. 

(**********************************
  BASIC SUBSTITUTION LEMMAS
 **********************************)

(* Substituting for itself *)
Lemma subst_gvar_idem : forall X p,
  [Gvar X//X]p = p.
Proof.
  inductions p; simpls*; fequals; cases_if*. congruence.
Qed.

(* Substituting into a GVariable *)
Lemma subst_gvar_equal : forall X q, [q//X](Gvar X) = q.
Proof.
  introv; unfolds; cases_if*.
Qed.

(* Substituting for itself *)
Lemma subst_lvar_idem : forall x p,
  [Lvar x /// x]p = p.
Proof.
  inductions p; simpls*; fequals; cases_if*. congruence.
Qed.

(* Substituting into an LVariable *)
Lemma subst_lvar_equal : forall x q, [q///x](Lvar x) = q.
Proof.
  introv; unfolds; cases_if*.
Qed.

(**********************************
  FRESHNESS WITH SWAP AND SUBST
 **********************************)

(* If X is fresh in a gsubstitution, then it is 
   fresh in the process being substituted into *)
Lemma gfresh_from_subst : forall X Y p q,  X <> Y -> X # ([q//Y]p) -> X # p.
Proof with freshen_up.
  induction p; intros; simpls*...
  cases_if*; subst...
Qed.

(* If X is fresh in a lsubstitution, then it is 
   fresh in the process being substituted into *)
Lemma gfresh_from_lsubst : forall X Y p x,
  X <> Y -> X # ([(Gvar Y) /// x]p) -> X # p.
Proof with freshen_up.
  induction p; intros q Hf1 Hf2; simpls... 
  cases_if*.
Qed.

(* If X is fresh in the process substituted into 
   and the process substituted for, then it is fresh 
   in the substitution itself *)
Lemma gfresh_in_subst_p_q : forall X Y p q,
  X # p -> X # q -> X # ([q // Y]p).
Proof with freshen_up.
  induction p; intros q Hf1 Hf2; simpls...
  cases_if*; subst...
Qed.

(* If X is fresh in the process substituted into 
   and the process substituted for, then it is fresh 
   in the substitution itself *)
Lemma gfresh_in_lsubst_p_q : forall X x p q,
  X # p -> X # q -> X # ([q /// x]p).
Proof with freshen_up.
  induction p; intros q Hf1 Hf2; simpls...
  cases_if*; subst...
  cases_if*.
Qed.

(* If X is fresh in the term being substituted for, 
   it is fresh in the resulting substitution *)
Lemma gfresh_in_self_subst : forall p X q, X # q -> X # [q // X]p.
Proof with freshen_up.
  inductions p; introv Hf; simpls*...
  cases_if*... 
Qed.

(* Substituting for a fresh GVariable has no effect *)
Lemma gfresh_subst_rewrite : forall (X:var) (p q:process),
  X # p -> [q // X]p = p.
Proof with freshen_up.
  introv H. inductions p; simpls*; try fequals... 
  cases_if*; subst; false...
Qed.

Lemma lfresh_subst_rewrite : forall (x:lvar) (p q:process),
  x ## p -> [q /// x]p = p.
Proof with freshen_up.
  introv H; inductions p; simpls; fequals...
  cases_if*; fequals; apply IHp. rewrite lfresh_receive in H. intuition eauto.
  cases_if*; subst; false...  
Qed.

(* If X is fresh in the term being substituted for, 
   it is fresh in the resulting substitution *)
Lemma lfresh_in_subst_p_q : forall x X p q, x ## p -> x ## q -> x ## ([q//X] p).
Proof with freshen_up.
  induction p; intros q Hfp Hfq; simpls~...
  rewrite lfresh_receive in *. inverts Hfp; jauto.
  cases_if*; subst; false...
Qed.

(* Swapping fresh variables has no effect *)
Lemma gfresh_swap_rewrite : forall X X' p,  
  X # p -> X' # p -> {{[X,X']}}p = p.
Proof with freshen_up.
  inductions p; simpls; intros; fequals... 
  unfolds; repeat cases_if*; false...
Qed.

(* Swapping preserves freshness *)
Lemma gfresh_swap_inv : forall X X' Y p,
  Y # p -> ({{X,X'}}Y) # ({{[X,X']}}p).
Proof with freshen_up.
  induction p; simpl; intros...
  unfold swap_gvar. repeat cases_if*; substs... 
Qed.

(* Substitution under freshness is swap *)
Lemma gfresh_subst_is_swap : forall (X Y:var) (p : process),
  Y # p -> [(Gvar Y)//X] p = {{[X,Y]}} p.
Proof with freshen_up.
  inductions p; simpls; introv  Hf; fequals; eauto; try apply IHm1; try apply IHm2...
  cases_if*; unfold swap_gvar; unfolds in Hf; repeat cases_if*. 
  apply notin_singleton in Hf; false~.
Qed.

(* Involution of substitution under freshness *)
Lemma gfresh_subst_involutive : forall p X Y, X # p ->
  [Gvar Y // X]([Gvar X // Y]p) = p.
Proof with freshen_up.
  inductions p; introv Hf; simpls; fequals*...
    cases_if; simpls; cases_if*.
      rewrite~ e. false... 
Qed.

Ltac fresh_solve :=
  try match goal with
    | [ |- ?X # ([?q // ?X]?p) ] => apply* (gfresh_in_self_subst p X q)
    | [ |- ?X # ([?q // ?Y]?p) ] => apply* (gfresh_in_subst_p_q X Y p q)
    | [ |- ?x ## ([?q // ?y]?p) ] => apply* (lfresh_in_subst_p_q x y p q)
    | [ |- ({{?X,?X'}}?Y) # ({{[?X,?X']}}?p) ] => apply* (gfresh_swap_inv X X' Y p)
    | [ H : ?X # ([?q // ?Y] ?p) |- ?X # ?p ] => apply* (gfresh_from_subst X Y p q)
    | [ H : ?X # ([?q // ?Y] ?p) |- _ ] => apply (gfresh_from_subst X Y p q) in H
    | [ H : ?X # ([(Gvar ?Y) /// ?x]?p) |- ?X # ?p ] => apply* (gfresh_from_lsubst X Y p x)
  end; freshen_up; try solve [intuition eauto].

(**********************************
  PROPERTIES OF E
 **********************************)

(* If X is fresh in a process, there can be no LVariables binding it *)
Lemma gfresh_E_empty : forall (X:var) (p:process), X # p -> (E X p) = \{}.
Proof with fresh_solve.
  intros X. induction p; simpl; intros H; auto...
  cases_if*; false... 
  rewrite IHp1... rewrite IHp2...
  rewrite~ union_empty_l.
Qed.

Lemma E_subst_inv : forall p l l' X Z, X <> Z -> (l \in E X p <-> l \in E X ([(Lvar l') // Z]p)).
Proof with fresh_solve.
  inductions p; introv Hneq; splits; simpls*; introv Hyp; try solve [rewrite~ IHp; eassumption]; try solve [rewrite~ <- IHp].
  repeat cases_if*...
  false n...   
  rewrite in_union in *; inverts Hyp.
    left; auto.
    right; rewrite~ <- IHp.
  repeat cases_if*...
  false n...
    rewrite in_union in *; inverts Hyp.
    left; auto.
    right; rewrite~ IHp; eassumption. 
  rewrite* in_empty in Hyp.
  cases_if*.
  rewrite in_union in *; inverts Hyp; [left | right].
  rewrite~ <- IHp1. rewrite~ <- IHp2.
  rewrite in_union in *; inverts Hyp; [left | right].
  rewrite~ IHp1; eassumption. rewrite~ IHp2; eassumption.
Qed. 

(**********************************
  DECOMPOSING SUBSTITUTION
 **********************************)

(* Expanding substitution under LVariable freshness *)
Lemma subst_decomposition_l : forall (x:lvar) (X:var) (p q:process),
  x##p -> ~(x \in (E X p)) -> [q//X]p = [q///x][Lvar x//X]p.
Proof with freshen_up.
  introv H H1.
  inductions p; simpls; fequals; repeat cases_if; fequals*.
  repeat rewrite~ gfresh_subst_rewrite.
  false H1. rewrite in_union. left. rewrite~ in_singleton.
  apply IHp. rewrite lfresh_receive in H. inversion~ H. false~ n. rewrite~ gfresh_E_empty.  
  apply IHp. rewrite lfresh_receive in H. inversion~ H. false~ n. rewrite~ in_union in H1.
  unfolds in H. apply notin_singleton in H. false~ H.
  rewrite~ subst_lvar_equal. 
  apply IHp1... rewrite~ in_union in H1.
  apply IHp2... rewrite~ in_union in H1.
Qed.

(* Expanding substitution under GVariable freshness *)
Lemma subst_decomposition_g : forall (X Y:var) (p q:process),
  Y#p -> [q//X]p = [q//Y][(Gvar Y)//X]p.
Proof with freshen_up.
  introv H.
  inductions p; simpls; fequals; repeat cases_if*; simpls*...
    cases_if~.
    cases_if~. false H. simpl. rewrite~ in_singleton.
Qed.

Lemma lsubst_decomposition_g : forall (x:lvar) (X:var) (p q:process),
  X#p -> [q///x]p = [q//X][Gvar X///x]p.
Proof.
  introv H.
  inductions p; simpls*; try rewrite* IHp; repeat cases_if*.
  rewrite~ gfresh_subst_rewrite. rewrite~ subst_gvar_equal.
  false. unfolds in H. apply notin_singleton in H. congruence.
  fequals;[ apply IHp1 | apply IHp2]; rewrite gfresh_par_iff in *; jauto.
Qed.

(**********************************
  COMMUTATIVITY OF SUBSTITUTION
 **********************************)

Lemma substs_commute_gg : forall (X Y:var) (p q r:process),
  X<>Y -> X#r -> Y#q -> [r//Y] ([q//X] p) = [q//X] ([r//Y] p).
Proof.
  introv Hneq Hf1 Hf2. induction p; simpls~; try rewrite~ IHp.
    cases_if~; cases_if~.
      rewrite~ gfresh_subst_rewrite. rewrite e. simpl. cases_if~.
      rewrite~ (gfresh_subst_rewrite X). rewrite e. simpl. cases_if~.
      simpls. cases_if; cases_if~.
    fequals.
Qed.

Lemma substs_commute_lg : forall (x:lvar) (X:var) (p q r:process),
  X#r -> x##q -> [r///x] ([q//X] p) = [q//X] ([r///x] p).
Proof.
  introv H H0.
  induction p; simpls*; try rewrite* IHp; repeat cases_if*.
  rewrite~ gfresh_subst_rewrite. rewrite~ lfresh_subst_rewrite.
  fequals~.
Qed. 

(* More interaction of two substitutions  *)
Lemma subst_on_subst_same : forall X (p q r:process),
  X#r -> [r//X] ([q//X] p) = [([r//X] q)//X] p.
Proof.
  inductions p; simpls; introv Hf; try solve [fequals*].
  repeat cases_if*.
  simpls; cases_if*.
Qed.

(**********************************
  MORΕ ON SWAPPING AND SUBSTITUTION
 **********************************)

(* Interaction of two substitutions  *)
Lemma subst_on_subst : forall (X Y:var) (p q r:process),
  X<>Y -> X#r -> [r//Y] ([q//X] p) = [([r//Y] q)//X] ([r//Y] p).
Proof.
  inductions p; simpls; introv Hneq Hf; try solve [fequals*].
  repeat cases_if*.
  rewrite e. rewrite~ subst_gvar_equal.
  rewrite (gfresh_subst_rewrite X r ([r//Y]q)).
  rewrite e. rewrite~ subst_gvar_equal. assumption.
  rewrite <- gfresh_gvar in n. rewrite <- gfresh_gvar in n0.
  repeat rewrite* gfresh_subst_rewrite.
Qed.

(* Interaction of swapping and substitution *)
Lemma swap_on_subst : forall (X X' Y:var) (p q:process),
  {{[X,X']}} ([q//Y] p) = [({{[X,X']}} q)//({{X,X'}} Y)] ({{[X,X']}} p).
Proof.
  inductions p; introv; simpl; fequals*; repeat cases_if*.
  false n. rewrite* swap_gvar_equal_iff.
Qed.  

(* Interaction of substitution and swapping *)
Lemma subst_on_swap : forall X X' Y p q, [q//Y]({{[X,X']}}p)= {{[X,X']}}([({{[X,X']}}q)//({{X,X'}}Y)]p).
Proof.
  introv. rewrite <- swap_involutive with (X:=X) (Y:=X') (p:= [q // Y]({{[X, X']}}p)).
  fequals. rewrite swap_on_subst.
  fequals. apply swap_involutive.
Qed.

(* Interaction of channel swapping and substitution *)
Lemma swap_chan_on_subst: forall m n p q X,
  (|m,n| ([q//X]p)) = [(|m,n|q) // X] (|m,n|p).
Proof.
  induction p;  introv; simpls~; try rewrite~ IHp.
    cases_if. cases_if~. cases_if~.
    rewrite IHp1. rewrite~ IHp2.
Qed.

(**********************************
  ON GVs AND LVs
 **********************************)

Lemma GV_subst: forall p q X, ~ X # p -> GV ([q // X]p) = GV p \- \{X} \u GV q.
Proof with freshen_up.
  induction p; introv Hlv; simpls~.
    false~ Hlv...

    cases_if; subst.
      rewrite remove_same; rewrite~ union_empty_l.
      false Hlv...
 
   (*lvar fresh is decidable*)
    rewrite gfresh_par_iff in Hlv; rew_logic in Hlv.
    tests : (X # p1); tests : (X # p2);
      try lets Hc1 : IHp1 q C; try lets Hc2 : IHp2 q C0;
      try rewrite Hc1; try rewrite Hc2; try rewrite~ gfresh_subst_rewrite.
    
    false*.
 
    rewrite union_assoc. rewrite <- remove_union_distr.
    unfold fresh_gvar in C. rewrite~ (remove_while_notin C).          
    rewrite <- remove_union_distr.
    unfold fresh_gvar in C0. rewrite~ (remove_while_notin C0).
    repeat rewrite <- union_assoc. rewrite~ (union_comm (GV q)).
    
    rewrite~ <- union_assoc. rewrite union_comm_assoc. rewrite union_comm.
    do 2 rewrite <- union_assoc. rewrite union_same. rewrite~ union_assoc.
    rewrite~ remove_union_distr.    

    false Hlv... 
Qed.

Lemma LV_subst: forall p q X, LV ([q // X]p) \c LV p \u LV q.
Proof with freshen_up.
  induction p; introv Hlv; simpls; unfold subset in *; 
  try solve [rewrite* in_union; cases_if*].
    applys* IHp.

    rewrite in_remove in Hlv; destruct Hlv as (Hlv1 & Hlv2).
    apply IHp in Hlv1.
    rewrite in_union in *; rewrite* in_remove.

    repeat rewrite in_union in *; inverts Hlv as Hlv; 
    [apply IHp1 in Hlv | apply IHp2 in Hlv]; rewrite* in_union in Hlv.
Qed.
