(*********************************************************************

  HOC20Barbed

**********************************************************************

  Copyright (C) 2014
  Simon Boulier, Martin Escarra, Petar Maksimovic, Alan Schmitt

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

**********************************************************************)

Require Import HOC01Defs.
Require Import HOC02BaseLemmas.
Require Import HOC03CanonicalLemmas.
Require Import HOC04DefLTS.
Require Import HOC05WFProcesses.
Require Import HOC06FreshLemmas.
Require Import HOC07SizeLemmas.
Require Import HOC08SubstLemmas.
Require Import HOC09GVLemmas.
Require Import HOC10CongrLemmas.
Require Import HOC11TransLemmas.
Require Import HOC12Guarded.
Require Import HOC13Bisimulations.
Require Import HOC14IObis.
Require Import HOC17NORbis.


Require Import Relations.


Inductive context : Set :=
| CSend : chan -> context -> context
| CAbs : chan -> var -> context -> context
| CParL : context -> wfprocess -> context
| CParR : wfprocess -> context -> context
| CHole : context.

Fixpoint fill (C : context) (p : wfprocess) : wfprocess :=
  match C with
    | CSend a C => wfp_Send a (fill C p)
    | CAbs a X C => wfp_Abs a X (fill C p)
    | CParL C q => wfp_Par (fill C p) q
    | CParR q C => wfp_Par q (fill C p)
    | CHole => p
  end.

Definition context_closed (R : relation wfprocess) : Prop :=
  forall p q, (R p q) -> forall C, R (fill C p) (fill C q).

Definition perform_output (p : wfprocess) (a:chan) : Prop :=
  exists (p' p'': wfprocess), {{p--LabOut a p''->>(AP p')}} .

Definition perform_input (p : wfprocess) (a:chan) : Prop :=
  exists f, {{p--LabIn a->>(AA f)}} .

Definition out_barb_preserving (R : relation wfprocess) : Prop :=
  forall p q,  (R p q) -> forall a, perform_output p a -> perform_output q a.

Definition in_barb_preserving (R:relation wfprocess) : Prop :=
  forall p q, (R p q) -> forall a, perform_input p a -> perform_input q a.

Definition async_barbed_relation R : Prop :=
  (Symmetric R) /\ (tau_relation R) /\ (context_closed R) /\ (out_barb_preserving R).

Definition async_barbed p q : Prop :=
  exists R, async_barbed_relation R /\ (R p q).

Instance async_barbed_sym : Symmetric async_barbed.
Proof.
  introv (R & (Hs & Ht & Ho & Hi) & Hr). exists* R; repeat splits*.
Qed.

Lemma async_barbed_tau : tau_relation async_barbed.
Proof.
  introv (R & (Hsym & Htau & Hcc & Houtb) & Hr) Ht.
  lets (q' & Htq & Hr') : Htau Hr Ht. 
  exists q'; splits~; exists R; repeat splits*.
Qed.

Lemma async_barbed_context_closed : context_closed async_barbed.
Proof.
  introv (R & (Hsym & Htau & Hcc & Houtb) & Hr).
  exists R; repeat splits*.
Qed.

Lemma async_barbed_out_barb_preserving : out_barb_preserving async_barbed.
Proof.
  introv (R & (Hsym & Htau & Hcc & Houtb) & Hr) Ht.
  eapply Houtb; eassumption.
Qed.

Inductive R_trans_ab : binary wfprocess :=
| Rtrans_ab : forall p q r, async_barbed p q -> async_barbed q r -> R_trans_ab p r.

Hint Resolve Rtrans_ab.

Instance async_barbed_trans : Transitive async_barbed.
Proof.
  introv Hxy Hyz. exists R_trans_ab; splits*; clear; repeat splits*.
  
  + introv Hr. inverts Hr. econstructor. 
      symmetry in H0. eassumption.
      symmetry in H. eassumption.

  + introv Hr Ht. inverts Hr.
    lets (q' & Htq' & Hab') : async_barbed_tau H Ht.
    lets (q'' & Htq'' & Hab'') : async_barbed_tau H0 Htq'.
    eexists. splits; try eassumption. econstructor; eassumption.

  + introv Hr; introv; inverts Hr.
    lets Htr' : async_barbed_context_closed H C.
    lets Htr'' : async_barbed_context_closed H0 C.
    econstructor; eassumption.

  + introv Hr Ht. inverts Hr.
    lets Htr' : async_barbed_out_barb_preserving H Ht.
    lets~ Htr'' : async_barbed_out_barb_preserving H0 Htr'.
Qed.

Theorem IObis_async_barbed_lr : forall p q,
  IObis p q -> async_barbed p q.
Proof.
  intros p q Hio.
  exists IObis. repeat splits. 
    apply IObis_sym.
    apply IObis_tau.
    clear p q Hio. intros p q Hio C.
    induction C; intros; simpl.
      applys~ IObis_congr_send. 
      applys~ IObis_congr_abs. 
      applys~ IObis_congr_par.
      applys~ IObis_congr_par. 
      assumption.
    clear p q Hio. intros p q Hio a H.
      destruct H as [p' [p'' H]]. eapply IObis_out in H; eauto.
      destruct H as [q' [q'' H]]. 
      exists~ q' q''. applys~ H. auto.
Qed.
